# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

from .adapters import MatrixAdapter, MatrixAdapterRow, Truncated2DMatrix, to_trunc, DupeList
from .base_formatter import unicode_dots


fixed_width_dots = {
    "h": "...",  # '⋯'
    "v": ":",  # '⋮'
    "d": "..."  # '⋱'
}
"""
Many fixed-width fonts don't have unicode ellipsis characters. Output viewed in those fonts may have misalignment
issues due to this.

This mapping uses plain characters present in all fonts that should be the same width everywhere.
"""

unicode_to_fixed_width = {
    u: fixed_width_dots[k] for k, u in unicode_dots.items()
}


def _single_line(s: str) -> str:
    lines = s.split("\n")
    for i, line in enumerate(lines):
        lines[i] = line.strip()
    return " ".join(lines)


class ListConverter:
    def __init__(self, max_rows, max_cols, num_after_dots, floatfmt, **_):
        super().__init__()
        self.max_rows = max_rows
        self.max_cols = max_cols
        self.num_after_dots = num_after_dots
        self.floatfmt = floatfmt
        self.empty_cell = ""
        self.replace_newline_char = " "
        self.use_fixed_width_dots = True

    def pprint(self, obj):
        if obj is None:
            return self.empty_cell

        if isinstance(obj, (int, float)):
            formatted = self.floatfmt(obj)
            return obj if repr(obj) == formatted else formatted

        if type(obj).__module__ == "numpy":
            # numpy is installed
            import numpy as np

            if np.isscalar(obj):
                py = obj.item()
                if isinstance(py, (int, float, complex, bool, str, bytes)):
                    obj = py  # fall through for possible further rendering
                else:
                    return _single_line(repr(obj))
            else:
                # do not fall through to the regular numpy driver below because of missing multiline support in tabular.
                return _single_line(str(obj))

        if isinstance(obj, complex):
            sign = "-" if obj.imag < 0 else "+"
            r = str(self.pprint(obj.real)).replace("e+", "e")
            c = str(self.pprint(obj.imag)).replace("e+", "e")
            if "e" in r:
                # add space to visually separate the exponent from the complex plus
                sign = f" {sign} "
            return f"{r}{sign}{c}i"

        if isinstance(obj, str):
            if self.use_fixed_width_dots and obj in unicode_to_fixed_width:
                return unicode_to_fixed_width[obj]
            if obj in unicode_dots.values():
                return obj
            return repr(obj)

        if isinstance(obj, DupeList):
            return [self.pprint(sub) for sub in obj]

        from . import _get_adapter
        adapter = _get_adapter(obj, None, unsupported_raise=False)
        if adapter:
            fmt = ListConverter(max_rows=max(self.max_rows / 2, 2),
                                max_cols=max(self.max_cols / 2, 2),
                                num_after_dots=0,
                                floatfmt=self.floatfmt)
            ret, _, _ = fmt.to_lists_and_labels(adapter)
            return ret

        return obj

    def _write_matrix(self, mat: MatrixAdapterRow, is_vector=False):
        if isinstance(mat, Truncated2DMatrix):
            mat.apply_dots(unicode_dots)

        nrows, ncols = mat.get_shape()

        ret = []

        # values
        for row_idx in range(nrows):
            row = [self.empty_cell] * ncols
            for col_idx, cell in mat.get_row(row_idx, col_range=(0, ncols)):
                row[col_idx] = self.pprint(cell)

            ret.append(row)
        return ret[0] if is_vector else ret

    def to_lists_and_labels(self, mat: MatrixAdapter, is_1d_ok=True):
        is_vector = is_1d_ok and len(mat.get_shape()) == 1
        if not isinstance(mat, MatrixAdapterRow) or \
                len(mat.get_shape()) != 2 or \
                mat.get_shape()[0] > self.max_rows or \
                mat.get_shape()[1] > self.max_cols:
            mat = to_trunc(mat, self.max_rows, self.max_cols, self.num_after_dots)

        return self._write_matrix(mat, is_vector=is_vector), mat.get_row_labels(), mat.get_col_labels()



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

import dataclasses
from dataclasses import dataclass, asdict
from shutil import get_terminal_size
from typing import Any, Type, Callable, Dict, List, Union, Iterable, Optional

from .adapters import Driver, MatrixAdapter
from .html_formatter import HTMLTableFormatter, NotebookHTMLFormatter
from .latex_formatter import LatexFormatter, python_scientific_to_latex_times10
from matrepr.html import load_ipython_extension, unload_ipython_extension


@dataclass
class MatReprParams:
    max_rows: int = 11
    """Maximum number of rows in output."""

    max_cols: int = 15
    """Maximum number of columns in output."""

    width_str: int = None
    """
    Maximum width of text output.
     - If 0 then autodetect terminal size.
     - If None then use max_cols.

    Some methods that print to the terminal may automatically set this value.
    """

    num_after_dots: Union[int, float] = 0.5
    """
    If a matrix has more rows or columns than allowed then an ellipsis (three dots) is emitted to cover the excess.
    This parameter controls how many rows/columns are drawn at the end of the matrix.
    
    For example, a value of 1 means the final row and final column are emitted in addition to the top-left corner.
    A value of 2 means the final two rows and columns are emitted, with a correspondingly smaller top-left corner.
    
    A fractional value between 0 and 1 means a fraction of `max_rows` and `max_cols`.
    
    Note: May be set to 0 for very large matrices without fast row/column indexing.
    """

    cell_align: str = "center"
    """Horizontal text alignment for non-zero values."""

    indices: bool = True
    """Whether to show row/column indices."""

    title: Union[bool, str] = True
    """Title of table. If `True` then generate matrix description such as dimensions, nnz, datatype."""

    title_latex: bool = False
    """If not None then overrides `title` for Latex output only."""

    latex_matrix_env: str = "bmatrix"
    """Latex environment to use for matrices. For Jupyter this should be one supported by MathJax."""

    latex_tensor_env: str = "matrix"
    """
    Latex environment to use for tensors rendered as a 2D matrix.
    For Jupyter this should be one supported by MathJax.
    """

    latex_dupe_matrix_env: str = "Bmatrix"
    """Latex environment to use when a cell has multiple elements."""

    precision: int = 4
    """
    Floating-point precision. May be overriden by `floatfmt`.
    """

    floatfmt: Union[str, Callable[[float], str]] = None
    """
    A format string to format numbers as string, usable by :code:`format()`.

    May also be a `Callable` that converts numbers to strings.
    
    If None then `precision` is used.
    """

    floatfmt_latex: Callable[[float], str] = None
    """
    Overwrites `floatfmt` for LaTeX output. If None then uses `floatfmt` but converts scientific
    notation: :raw:`1e22` becomes :raw:`1 \\times 10^{22}`
    """

    row_labels: Iterable[str] = None
    """Labels for matrix rows. If None then use row index."""

    col_labels: Iterable[str] = None
    """Labels for matrix columns. If None then use column index."""

    def set_precision(self, precision, g=True):
        """
        Precision to use for floating-point to string conversion.

        :param precision: floating-point precision
        :param g: Whether to use :code:`g` or :code:`f` formatting.
        """
        fmt_str = f".{precision}{'g' if g else ''}"
        self.floatfmt = lambda f: format(f, fmt_str)

    def _assert_one_of(self, var, choices):
        if getattr(self, var) not in choices:
            raise ValueError(f"{var} must be one of: " + ", ".join(choices))

    def get(self, **kwargs):
        ret = dataclasses.replace(self)

        # Allow some explicit overwrites for convenience
        if "title" in kwargs and "title_latex" not in kwargs:
            kwargs["title_latex"] = kwargs["title"]

        # Update all parameters with the ones in kwargs
        for key, value in kwargs.items():
            if hasattr(ret, key):
                setattr(ret, key, value)

        # Handy type conversions
        if ret.floatfmt is None:
            ret.set_precision(ret.precision)
        elif isinstance(ret.floatfmt, str):
            fmt_str = ret.floatfmt
            ret.floatfmt = lambda f: format(f, fmt_str)

        # validate
        ret._assert_one_of("cell_align", ['center', 'left', 'right'])
        ret.max_rows = max(1, ret.max_rows)
        ret.max_cols = max(1, ret.max_cols)

        # Apply some default rules
        if ret.title_latex is None:
            ret.title_latex = ret.title

        if ret.floatfmt_latex is None:
            ret.floatfmt_latex = lambda f: python_scientific_to_latex_times10(ret.floatfmt(f))

        # compute automatic values
        if ret.width_str == 0 or (ret.width_str is None and kwargs.get("auto_width_str", False)):
            width_str, _ = get_terminal_size()
            ret.width_str = width_str

        return ret

    def to_kwargs(self):
        return asdict(self)


params = MatReprParams()
"""
Changeable default parameters that apply unless overridden in a method call.
"""
_drivers: List[Type[Driver]] = []
_driver_map: Dict[str, Type[Driver]] = {}


def register_driver(driver: Type[Driver]):
    _drivers.append(driver)

    for type_str, _ in driver.get_supported_types():
        _driver_map[type_str] = driver


def _register_bundled():
    """
    Register the built-in drivers.
    """
    from .adapters.scipy_driver import SciPyDriver
    register_driver(SciPyDriver)

    from .adapters.graphblas_driver import GraphBLASDriver
    register_driver(GraphBLASDriver)

    from .adapters.sparse_driver import PyDataSparseDriver
    register_driver(PyDataSparseDriver)

    from .adapters.numpy_driver import NumpyDriver
    register_driver(NumpyDriver)

    from .adapters.list_like import ListDriver
    register_driver(ListDriver)


_register_bundled()


def _get_driver(mat, unsupported_raise=True):
    type_str = ".".join((type(mat).__module__, type(mat).__name__))
    driver = _driver_map.get(type_str, None)
    if not driver and unsupported_raise:
        # print("Supported types: \n" + "\n".join(sorted(list(_driver_map.keys()))))
        raise AttributeError("Unsupported type: " + type_str)
    return driver


def _get_adapter(mat: Any, options: Optional[MatReprParams], unsupported_raise=True) -> MatrixAdapter:
    if isinstance(mat, MatrixAdapter):
        adapter = mat
    else:
        driver = _get_driver(mat, unsupported_raise=unsupported_raise)
        adapter = driver.adapt(mat) if driver else None
        if not adapter and unsupported_raise:
            raise AttributeError("Unsupported matrix")

    if options and options.row_labels:
        adapter.row_labels = options.row_labels
    if options and options.col_labels:
        adapter.col_labels = options.col_labels
    return adapter


def to_html(mat: Any, notebook=False, **kwargs) -> str:
    """
    Render a matrix to HTML.

    :param mat: A supported matrix.
    :param notebook: If true then adds extra styling appropriate for display in a Jupyter notebook.
    :param kwargs: Any argument in :class:`MatReprParams`.
    :rtype: str
    :return: A string containing an HTML representation of `mat`.
    """
    options = params.get(**kwargs)
    adapter = _get_adapter(mat, options)

    if notebook:
        formatter = NotebookHTMLFormatter(**options.to_kwargs())
    else:
        formatter = HTMLTableFormatter(**options.to_kwargs())

    return str(formatter.format(adapter))


def to_latex(mat: Any, **kwargs):
    """
    Render a matrix to LaTeX.

    :param mat: A supported matrix.
    :param kwargs: Any argument in :class:`MatReprParams`.
    :rtype: str
    :return: A string containing a LaTeX representation of `mat`.
    """
    options = params.get(**kwargs)
    adapter = _get_adapter(mat, options)

    formatter = LatexFormatter(**options.to_kwargs())

    return str(formatter.format(adapter))


def to_str(mat: Any, **kwargs) -> str:
    """
    Render a matrix to a string.

    Internally uses `tabulate` and will pass through any arguments not explicitly set by this method.

    :param mat: A supported matrix.
    :param kwargs: Any argument in :class:`MatReprParams`.
    :return: a string representation of `mat`.
    """
    options = params.get(**kwargs)

    ret = []

    if options.title:
        if options.title is True:
            adapter = _get_adapter(mat, None)
            title = adapter.describe()
        else:
            title = options.title
        ret.append(title)

    from .string_formatter import to_tabulate
    ret.append(to_tabulate(mat, **kwargs))

    return "\n".join(ret)


def mprint(mat: Any, **kwargs):
    """
    Prints the output of :func:`to_str`.

    :param mat: A supported matrix.
    :param kwargs: Any argument in :class:`MatReprParams`.
    """
    print(to_str(mat, auto_width_str=True, **kwargs))


def mdisplay(mat: Any, method="html", **kwargs):
    """
    Display a matrix in Jupyter.

    :param mat: A supported matrix.
    :param method: Style to use. One of :code:`"html"`, :code:`"latex"`, :code:`"str"`.
    :param kwargs: Any argument in :class:`MatReprParams`.
    """
    from IPython.display import display, HTML, Latex, Pretty

    if method == "html":
        display(HTML(to_html(mat, notebook=True, **kwargs)))
    elif method == "latex":
        display(Latex('$' + to_latex(mat, **kwargs) + '$'))
    elif method == "str":
        display(Pretty(to_str(mat, auto_width_str=True, **kwargs)))
    else:
        raise ValueError("Unknown method: " + method)


def _register_jupyter_formatter(ipython, mime_type: str, repr_method: Optional[Callable]):
    """
    See https://ipython.readthedocs.io/en/stable/config/integrating.html
    """
    if not ipython:
        return

    formatter = ipython.get_ipython().display_formatter.formatters[mime_type]

    for driver in _drivers:
        for type_str, register_with_jupyter in driver.get_supported_types():
            if register_with_jupyter:
                if repr_method:
                    formatter.for_type(type_str, repr_method)
                else:
                    formatter.pop(type_str)


__all__ = ["to_html", "to_latex", "to_str", "mprint", "mdisplay"]



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

from typing import Any, Iterable, Tuple

import sparse
from sparse import COO

from . import describe, MatrixAdapterRow, MatrixAdapterCoo, TensorAdapterCooRow


def sizeof_fmt(num, suffix="B"):
    for unit in ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"):
        if abs(num) < 1024.0:
            return f"{num:3.1f}".rstrip('0').rstrip('.')+f"{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"


class PyDataSparseBase:
    def __init__(self, mat):
        self.mat = mat

    def get_shape(self) -> tuple:
        return self.mat.shape

    def describe(self) -> str:
        parts = [
            self.mat.format,
            f"fill_value={self.mat.fill_value}",
        ]

        if not hasattr(self.mat, "__setitem__"):
            parts.append("read-only")

        if hasattr(self.mat, "nbytes"):
            parts.append(sizeof_fmt(self.mat.nbytes))

        return describe(shape=self.mat.shape,
                        nnz=self.mat.nnz, nz_type=self.mat.dtype,
                        notes=", ".join(parts))


class PyDataSparse1DAdapter(PyDataSparseBase, MatrixAdapterRow):
    def __init__(self, mat):
        assert mat.ndim <= 1
        super().__init__(mat)

    def get_row(self, row_idx: int, col_range: Tuple[int, int]) -> Iterable[Any]:
        assert row_idx == 0

        ret = COO(self.mat[slice(*col_range)])
        ret.coords += col_range[0]
        return zip(ret.coords.flat, ret.data.flat)


class PyDataSparse2DAdapter(PyDataSparseBase, MatrixAdapterCoo):
    def __init__(self, mat):
        assert mat.ndim == 2
        super().__init__(mat)

    def get_coo(self, row_range: Tuple[int, int], col_range: Tuple[int, int]) -> Iterable[Tuple[int, int, Any]]:
        ret = COO(self.mat[slice(*row_range), slice(*col_range)])
        ret.coords[0] += row_range[0]
        ret.coords[1] += col_range[0]
        return zip(ret.coords[0], ret.coords[1], ret.data)


class PyDataSparseNDAdapter(PyDataSparseBase, TensorAdapterCooRow):
    def __init__(self, mat):
        super().__init__(mat)
        self.coo = sparse.as_coo(mat)

    def get_shape(self) -> tuple:
        return self.mat.nnz, self.mat.ndim+1

    def get_dense_row(self, row_idx: int, col_range: Tuple[int, int]) -> Iterable[Any]:
        ret = self.coo.coords[:, row_idx]
        ret = list(ret) + [self.coo.data[row_idx]]
        return ret[slice(*col_range)]


def _warmup_indexing():
    _ = COO(coords=[[0, 1], [2, 3]], data=[111, 222], shape=(5, 5))[:1]


_warmup_indexing()



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

from typing import Any, Iterable, Tuple
from . import describe, Driver, MatrixAdapterRow, MatrixAdapterCol


def is_single(x):
    return (not hasattr(x, '__len__') or isinstance(x, str))\
        or type(x).__module__.startswith("scipy")\
        or type(x).__module__.startswith("sparse")\
        or type(x).__name__ == "Matrix"


def count_none(iterable):
    return sum((1 if x is None else 0) for x in iterable)


class ListAdapter(MatrixAdapterRow):
    def __init__(self, mat: list):
        self.mat = mat
        self.row_lengths = [(1 if is_single(row) else len(row)) for row in mat]
        if len(mat) == 0 or all(is_single(row) for row in mat):
            # 1D
            self.shape = (len(mat),)
        else:
            # 2D
            self.shape = (len(mat), max(self.row_lengths if self.row_lengths else [0]))
        self.nnz = sum((1 if is_single(row) else len(row) - count_none(row)) for row in mat)

    def get_shape(self) -> Tuple[int, int]:
        return self.shape

    def describe(self) -> str:
        return describe(shape=self.shape, nnz=self.nnz, notes=None)

    def get_row(self, row_idx: int, col_range: Tuple[int, int]) -> Iterable[Tuple[int, Any]]:
        row = self.mat if len(self.shape) == 1 else self.mat[row_idx]

        if is_single(row):
            # this row is a single element
            if col_range[0] <= 0 < col_range[1]:
                return ((0, row),)
            else:
                return ()
        else:
            return enumerate(row[col_range[0]:col_range[1]], start=col_range[0])


class List1DColumnAdapter(MatrixAdapterCol):
    def __init__(self, mat: list):
        self.mat = mat
        self.shape = (len(mat), 1)
        self.nnz = len(mat)

    def get_shape(self) -> tuple:
        return self.shape

    def describe(self) -> str:
        return describe(shape=self.shape, nnz=self.nnz, notes=None)

    def get_col(self, col_idx: int, row_range: Tuple[int, int]) -> Iterable[Tuple[int, Any]]:
        assert col_idx == 0
        return enumerate(self.mat[row_range[0]:row_range[1]], start=row_range[0])


class ListDriver(Driver):
    @staticmethod
    def get_supported_types() -> Iterable[Tuple[str, bool]]:
        return [
            ("builtins.list", False),
            ("builtins.tuple", False),
        ]

    @staticmethod
    def adapt(mat: Any):
        if isinstance(mat, (list, tuple)):
            return ListAdapter(mat)
        raise ValueError



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

import unittest

from matrepr import to_html, to_latex, to_str
import matrepr


class ListLikeTests(unittest.TestCase):
    def setUp(self):
        self.mats = [
            [],
            [1, 2, 3, 4],
            (1, 2, 3, 4),
            [[1, 2], [1003, 1004]],
            [[1, 2], [1003, 1004, 1005]],
        ]

    def test_no_crash(self):
        for mat in self.mats:
            res = to_html(mat, title=True)
            self.assertGreater(len(res), 10)

            res = to_latex(mat, title=True)
            self.assertGreater(len(res), 10)

    def test_no_crash_edge_cases(self):
        try:
            # noinspection PyUnresolvedReferences
            from scipy.sparse import coo_matrix
            # noinspection PyUnresolvedReferences
            import numpy as np

            sps_mat = coo_matrix(([1, 2, 3, 4], ([0, 0, 1, 1], [0, 1, 0, 1])), shape=(2, 2))

            dtype_a = np.dtype([("x", np.bool_), ("y", np.int64)], align=True)
            np_a = np.array([(False, 2)], dtype=dtype_a)[0]
            dtype_b = np.dtype("(3,)uint16")
            np_b = np.array([(1, 2, 3)], dtype=dtype_b)[0]
            np_2d = np.array([[11, 22], [33, 44]])
        except ImportError:
            sps_mat = np_a = np_b = np_2d = None

        list_mat = [
            (0, 12e34, 1e-6, None, 123456789),
            1,
            [complex(1, 2), complex(123456, 0.123456)],
            [[1], sps_mat, [2.1, 2.2], [[1.1, 2.2], [3.3, 4.4]]],
            ["multiline\nstring", "<escape!>", "\\begin{escape!}", {"a Python set"}],
            [np_a, np_b, np_2d]
        ]

        row_labels = ["scientific", "singleton", "complex", "nested", "strings", "numpy"]
        col_labels = ["one", "two", "three", "four", "five"]

        for max_cols in [3, 99]:
            res = to_html(list_mat, title=True, max_cols=max_cols, row_labels=row_labels, col_labels=col_labels)
            self.assertGreater(len(res), 10)

            res = to_latex(list_mat, title=True, max_cols=max_cols, row_labels=row_labels, col_labels=col_labels)
            self.assertGreater(len(res), 10)

            # tabulate SEPARATING_LINE detection will compare an element to a string. If that element happens to be a
            # numpy array, numpy issues this warning. It's a ValueError in future versions.
            # Ensure it's not thrown.
            import warnings
            with warnings.catch_warnings():
                warnings.filterwarnings(action='error', category=FutureWarning)

                res = to_str(list_mat, title=True, max_cols=max_cols, row_labels=row_labels, col_labels=col_labels)

            self.assertGreater(len(res), 10)

    def test_no_crash_edge_cases_2(self):
        try:
            import scipy.sparse
            import numpy as np

            mat = [
                scipy.sparse.random(3, 3, density=0.6),
                np.array([[1, 2], [3, 4]])
            ]

            res = to_html(mat)
            self.assertGreater(len(res), 10)
        except ImportError:
            pass

    def test_shape(self):
        mat = (1, 2, 3, 4)
        adapter = matrepr._get_adapter(mat, None)
        self.assertEqual((4,), adapter.get_shape())

        mat = [[1, 2], [1003, 1004, 1005]]
        adapter = matrepr._get_adapter(mat, None)
        self.assertEqual((2, 3), adapter.get_shape())


if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

from typing import Any, Iterable, Tuple

from . import Driver


class SciPyDriver(Driver):
    @staticmethod
    def get_supported_types() -> Iterable[Tuple[str, bool]]:
        formats = ("coo", "csr", "csc", "lil", "dok", "dia", "bsr")
        ret = []
        for mod in ("scipy.sparse", "scipy.sparse._arrays"):
            for arr_mat in ("array", "matrix"):
                for fmt in formats:
                    ret.append((f"{mod}.{fmt}_{arr_mat}", True))
        for fmt in formats:
            for arr_mat in ("array", "matrix"):
                ret.append((f"scipy.sparse._{fmt}.{fmt}_{arr_mat}", True))
            ret.append((f"scipy.sparse.{fmt}.{fmt}_matrix", True))
        return ret

    @staticmethod
    def adapt(mat: Any):
        import scipy.sparse
        from .scipy_impl import SciPyCSRAdapter, SciPyCSCAdapter, SciPyCOOAdapter

        if scipy.sparse.isspmatrix_csr(mat):
            return SciPyCSRAdapter(mat)
        if scipy.sparse.isspmatrix_csc(mat):
            return SciPyCSCAdapter(mat)
        else:
            return SciPyCOOAdapter(mat.tocoo())



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

from typing import Any, Iterable, Tuple

from . import Driver


class GraphBLASDriver(Driver):
    @staticmethod
    def get_supported_types() -> Iterable[Tuple[str, bool]]:
        return [
            ("graphblas.Matrix", True),
            ("graphblas.core.matrix.Matrix", True),
            ("graphblas.Vector", True),
            ("graphblas.core.vector.Vector", True),
        ]

    @staticmethod
    def adapt(mat: Any):
        from .graphblas_impl import GraphBLASMatrixAdapter, GraphBLASVectorAdapter
        if "Matrix" in str(type(mat)):
            return GraphBLASMatrixAdapter(mat)
        elif "Vector" in str(type(mat)):
            return GraphBLASVectorAdapter(mat)
        else:
            raise ValueError("Unknown type: " + str(type(mat)))



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

import unittest

import numpy.random
import scipy.sparse

from matrepr import to_html, to_latex, to_str
from .test_html import generate_fixed_value

numpy.random.seed(123)


class SciPyTests(unittest.TestCase):
    def setUp(self):
        self.mats = [
            scipy.sparse.coo_matrix(([], ([], [])), shape=(0, 0)),
            scipy.sparse.random(10, 10, density=0.4).tocoo(),
            scipy.sparse.random(5, 10, density=0.4).tocsr(),
            scipy.sparse.random(5, 1, density=0.4).tocsc(),
            scipy.sparse.coo_matrix(([], ([], [])), shape=(10, 10)).tocoo(),
            scipy.sparse.coo_matrix(([], ([], [])), shape=(10, 10)).tocsr(),
            scipy.sparse.coo_matrix(([], ([], [])), shape=(10, 10)).tocsc(),
            generate_fixed_value(10, 10),
            scipy.sparse.coo_matrix(([111, 222], ([0, 0], [0, 0])), shape=(13, 13))
        ]

    def test_no_crash(self):
        for mat in self.mats:
            res = to_html(mat)
            self.assertGreater(len(res), 10)

            res = to_latex(mat)
            self.assertGreater(len(res), 10)

            res = to_str(mat)
            self.assertGreater(len(res), 10)

    def test_formats(self):
        to_html_args = dict(notebook=False, max_rows=20, max_cols=20, title=False)

        expected = [to_html(m, **to_html_args) for m in self.mats]

        for fmt in ["coo", "csr", "csc", "dok", "lil"]:
            for i, source_mat in enumerate(self.mats):
                if 13 in source_mat.shape:
                    # this matrix has dupes, which act differently in different formats
                    continue
                mat = source_mat.asformat(fmt)
                res = to_html(mat, **to_html_args)
                self.assertEqual(expected[i], res)


class PatchSciPyTests(unittest.TestCase):
    def test_patch_scipy(self):
        source_mat = scipy.sparse.coo_matrix(([111, 222], ([0, 1], [0, 1])), shape=(10, 10))

        # noinspection PyUnresolvedReferences
        import matrepr.patch.scipy

        for fmt in ["coo", "csr", "csc", "dok", "lil"]:
            mat = source_mat.asformat(fmt)
            res = repr(mat)
            self.assertIn("222", res)
            self.assertIn("┌", res)  # a character used by MatRepr


if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

import re

from .adapters import MatrixAdapter, MatrixAdapterRow, Truncated2DMatrix, to_trunc, DupeList
from .base_formatter import BaseFormatter, unicode_dots


def python_scientific_to_latex_times10(s):
    return re.sub(r'e\+?(-?)0*([0-9]+)', ' \\\\times 10^{\\1\\2}', s)


def tex_escape(text):
    """
    :param text: a plain text message
    :return: the message escaped to appear correctly in LaTeX
    """
    conv = {
        '&': r'\&',
        '%': r'\%',
        '$': r'\$',
        '#': r'\#',
        '_': r'\_',
        '{': r'\{',
        '}': r'\}',
        '~': r'\textasciitilde{}',
        '^': r'\^{}',
        # '\\': r'\textbackslash{}',
        '\\': r'\\',
        # '<': r'\textless{}',
        # '>': r'\textgreater{}',
        " '": r' `',
    }
    regex = re.compile('|'.join(re.escape(str(key)) for key in sorted(conv.keys(), key=lambda item: - len(item))))
    return regex.sub(lambda match: conv[match.group()], text)


latex_dots = {
    "h": "\\dots",  # '⋯'
    "v": "\\vdots",  # '⋮'
    "d": "\\ddots"  # '⋱'
}

unicode_to_latex = {
    u: latex_dots[k] for k, u in unicode_dots.items()
}


class LatexFormatter(BaseFormatter):
    def __init__(self, max_rows, max_cols, num_after_dots, title_latex, latex_matrix_env, latex_tensor_env,
                 floatfmt_latex=None, latex_dupe_matrix_env="Bmatrix", **_):
        super().__init__()
        self.max_rows = max_rows
        self.max_cols = max_cols
        self.num_after_dots = num_after_dots
        self.title = title_latex
        self.latex_matrix_env = latex_matrix_env
        self.latex_tensor_env = latex_tensor_env
        self.dupe_env = latex_dupe_matrix_env
        self.floatfmt = floatfmt_latex
        if not self.floatfmt:
            self.floatfmt = lambda f: python_scientific_to_latex_times10(format(f))
        self.indent_width = 4

    def pprint(self, obj):
        if obj is None:
            return ""

        if isinstance(obj, (int, float)):
            return self.floatfmt(obj)

        if isinstance(obj, complex):
            r = obj.real
            c = obj.imag
            sign = "-" if c < 0 else "+"
            return f"{self.pprint(r)}{sign}{self.pprint(c)}i"

        dupe_list_env = self.dupe_env
        if isinstance(obj, str):
            if obj in unicode_to_latex:
                return unicode_to_latex[obj]
            elif "\n" in obj:
                obj = DupeList(obj.split("\n"))
                dupe_list_env = "matrix"
            else:
                return "\\textrm{" + tex_escape(obj) + "}"

        if isinstance(obj, DupeList):
            fmt = LatexFormatter(max_rows=len(obj), max_cols=1, num_after_dots=0, title_latex=None,
                                 latex_matrix_env=dupe_list_env,
                                 latex_dupe_matrix_env=self.dupe_env,
                                 latex_tensor_env=self.latex_tensor_env,
                                 floatfmt_latex=self.floatfmt)
            from .adapters.list_like import List1DColumnAdapter
            return str(fmt.format(List1DColumnAdapter(obj)))

        from . import _get_adapter
        adapter = _get_adapter(obj, None, unsupported_raise=False)
        if adapter:
            fmt = LatexFormatter(max_rows=max(self.max_rows / 2, 2),
                                 max_cols=max(self.max_cols / 2, 2),
                                 num_after_dots=0, title_latex=None,
                                 latex_matrix_env=self.latex_matrix_env,
                                 latex_tensor_env=self.latex_tensor_env,
                                 latex_dupe_matrix_env=self.dupe_env,
                                 floatfmt_latex=self.floatfmt)
            return str(fmt.format(adapter))

        return "\\textrm{" + tex_escape(str(obj)) + "}"

    def _write_matrix(self, mat: MatrixAdapterRow, matrix_env, indent: int = 0):
        if isinstance(mat, Truncated2DMatrix):
            mat.apply_dots(unicode_dots)

        nrows, ncols = mat.get_shape()
        self.write("\\begin{" + matrix_env + "}", indent=indent)

        body_indent = indent + self.indent_width

        # values
        for row_idx in range(nrows):
            row_contents = []

            col_range = (0, ncols)
            for col_idx, cell in enumerate(mat.get_dense_row(row_idx, col_range=col_range)):
                if cell is not None:
                    row_contents.append(self.pprint(cell))

                if col_idx == col_range[1] - 1:
                    if row_idx != nrows - 1:
                        row_contents.append("\\\\")
                else:
                    row_contents.append("&")

            self.write(" ".join(row_contents), body_indent)

        self.write("\\end{" + matrix_env + "}", indent=indent)

    def format(self, mat: MatrixAdapter):
        if self.title:
            title = mat.describe() if self.title is True else self.title
            self.write("\\stackrel{\\textrm{" + tex_escape(title) + "}}{")

        matrix_env = self.latex_tensor_env if mat.is_tensor() else self.latex_matrix_env

        if not isinstance(mat, MatrixAdapterRow) or \
                len(mat.get_shape()) != 2 or \
                mat.get_shape()[0] > self.max_rows or \
                mat.get_shape()[1] > self.max_cols:
            mat = to_trunc(mat, self.max_rows, self.max_cols, self.num_after_dots)

        self._write_matrix(mat, matrix_env)

        if self.title:
            self.write("}")
        return self



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

from typing import Any, Iterable, Tuple

from . import Driver


class NumpyDriver(Driver):
    @staticmethod
    def get_supported_types() -> Iterable[Tuple[str, bool]]:
        return [
            ("numpy.ndarray", True),
        ]

    @staticmethod
    def adapt(mat: Any):
        from .numpy_impl import NumpyArrayAdapter
        return NumpyArrayAdapter(mat)



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause
"""
Import this module to automatically print SciPy sparse matrices using MatRepr.

A `__repr__` method that calls MatRepr is inserted to all SciPy sparse matrix classes.

Also includes all public methods from the matrepr module for convenient one-line imports.
"""

from .. import *

import scipy.sparse


def _str_(mat):
    from matrepr.adapters.scipy_driver import SciPyDriver
    # Enable terminal width detection
    return to_str(SciPyDriver.adapt(mat), width_str=0, max_cols=9999)


try:
    scipy.sparse.bsr_matrix.__repr__ = _str_
    scipy.sparse.coo_matrix.__repr__ = _str_
    scipy.sparse.csc_matrix.__repr__ = _str_
    scipy.sparse.csr_matrix.__repr__ = _str_
    scipy.sparse.dia_matrix.__repr__ = _str_
    scipy.sparse.dok_matrix.__repr__ = _str_
    scipy.sparse.lil_matrix.__repr__ = _str_
except AttributeError:
    pass

try:
    scipy.sparse.bsr_array.__repr__ = _str_
    scipy.sparse.coo_array.__repr__ = _str_
    scipy.sparse.csc_array.__repr__ = _str_
    scipy.sparse.csr_array.__repr__ = _str_
    scipy.sparse.dia_array.__repr__ = _str_
    scipy.sparse.dok_array.__repr__ = _str_
    scipy.sparse.lil_array.__repr__ = _str_
except AttributeError:
    pass



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

from typing import Any, Iterable, Tuple

from . import Driver


class PyDataSparseDriver(Driver):
    @staticmethod
    def get_supported_types() -> Iterable[Tuple[str, bool]]:
        return [
            ("sparse.COO", True),
            ("sparse._coo.core.COO", True),
            ("sparse.DOK", True),
            ("sparse._dok.DOK", True),
            ("sparse.GCXS", True),
            ("sparse._compressed.compressed.GCXS", True),
            ("sparse.CSR", True),
            ("sparse._compressed.compressed.CSR", True),
            ("sparse.CSC", True),
            ("sparse._compressed.compressed.CSC", True),
        ]

    @staticmethod
    def adapt(mat: Any):
        from .sparse_impl import PyDataSparse1DAdapter, PyDataSparse2DAdapter, PyDataSparseNDAdapter
        dims = len(mat.shape)
        if dims < 2:
            return PyDataSparse1DAdapter(mat)
        elif dims == 2:
            return PyDataSparse2DAdapter(mat)
        else:
            return PyDataSparseNDAdapter(mat)



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

import unittest

import scipy.sparse

from matrepr import to_str
from matrepr.string_formatter import max_line_width
from .test_html import generate_fixed_value


class ToHTMLTests(unittest.TestCase):
    def setUp(self):
        self.mats = [
            scipy.sparse.random(10, 10, density=0.4),
            scipy.sparse.random(5, 10, density=0.4),
            scipy.sparse.random(5, 1, density=0.4),
            scipy.sparse.coo_matrix(([], ([], [])), shape=(10, 10)),
            generate_fixed_value(10, 10)
        ]

    def test_contents(self):
        mat = generate_fixed_value(9, 9)
        res = to_str(mat, max_rows=20, max_cols=20, title=True, indices=True)
        for value in mat.data:
            self.assertIn(f"{value}", res)
        for idx in range(10):
            self.assertIn(f"{idx}", res)

    def test_truncate(self):
        for fmt in ["csr", "csc", "coo"]:
            mat = generate_fixed_value(9, 9).asformat(fmt)

            for after_dots, expected_count in [
                (0, 25),  # 5*5
                (1, 25),  # 4*4 + 4 + 4 + 1
                (2, 25),  # 3*3 + 2*3 + 3*2 + 2*2
            ]:
                res = to_str(mat, max_rows=6, max_cols=6, num_after_dots=after_dots)
                count = 0
                for value in mat.data:
                    if f"{value}" in res:
                        count += 1
                self.assertEqual(expected_count, count)

    def test_width(self):
        mat = generate_fixed_value(9, 9)

        res = to_str(mat, max_cols=20, width_str=100, title=False, indices=False)
        self.assertLessEqual(max_line_width(res), 100)

        res = to_str(mat, max_cols=20, width_str=30, title=False, indices=False)
        self.assertLessEqual(max_line_width(res), 30)

        res = to_str(mat, max_cols=20, width_str=15, title=False, indices=True)
        self.assertLessEqual(max_line_width(res), 15)

        # test limit by max_cols
        res = to_str(mat, max_cols=2, width_str=200, title=False, indices=False)
        self.assertLessEqual(max_line_width(res), 20)

        # no crash test
        to_str(mat, max_cols=20, width_str=1, title=False, indices=True)
        to_str(mat, max_cols=20, width_str=0, title=False, indices=True)
        to_str(mat, max_cols=20, width_str=-1, title=False, indices=True)

    def test_title(self):
        mat = generate_fixed_value(4, 4)
        off = to_str(mat, title=False)
        self.assertNotIn("elements", off)

        on = to_str(mat, title=True)
        self.assertIn(f"{mat.nnz} 'int64' elements", on)

        title = "test title"
        custom = to_str(mat, title=title)
        self.assertIn(title, custom)


if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

"""
Removed in favor of Jupyter extension:

%load_ext matrepr.html
"""

raise DeprecationWarning("Please switch to Jupyter extension: %load_ext matrepr.html")



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

unicode_dots = {
    "h": "⋯",
    "v": "⋮",
    "d": "⋱",
}


class BaseFormatter:
    def __init__(self):
        self.lines = []

    def write(self, s, indent: int = 0):
        self.lines.append(" " * indent + s)

    def __str__(self):
        return "\n".join(self.lines)



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

"""
Jupyter extension that registers a LaTeX formatter for supported matrix types.

Usage:
%load_ext matrepr.latex
"""
from . import to_latex, _register_jupyter_formatter


def _repr_latex_(mat):
    return to_latex(mat)


def load_ipython_extension(ipython):
    _register_jupyter_formatter(ipython, "text/latex", _repr_latex_)


def unload_ipython_extension(ipython):
    _register_jupyter_formatter(ipython, "text/latex", None)



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause
"""
Import this module to automatically print python-graphblas `Matrix` and `Vector` using MatRepr.

A `__repr__` method that calls MatRepr is inserted to those classes.

Also includes all public methods from the matrepr module for convenient one-line imports.
"""

from .. import *

import graphblas


def _str_(mat):
    from matrepr.adapters.graphblas_driver import GraphBLASDriver
    # Enable terminal width detection
    return to_str(GraphBLASDriver.adapt(mat), width_str=0, max_cols=9999)


graphblas.Matrix.__repr__ = _str_
graphblas.Vector.__repr__ = _str_



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

import unittest

from matrepr import to_html, to_latex
import matrepr

from .test_scipy import generate_fixed_value

try:
    import graphblas as gb

    # Context initialization must happen before any other imports
    gb.init("suitesparse", blocking=True)

    have_gb = True
except ImportError:
    have_gb = False
    gb = None


@unittest.skipIf(not have_gb, "python-graphblas not installed")
class GraphBLASMatrixTests(unittest.TestCase):
    def setUp(self):
        self.mats = [
            gb.Matrix.from_coo([0, 1, 2, 3, 4], [0, 1, 2, 3, 4], [0, 1, 2, 3, 4], nrows=5, ncols=5),
        ]

    def test_no_crash(self):
        for mat in self.mats:
            res = to_html(mat, title=True)
            self.assertGreater(len(res), 10)

            res = to_latex(mat, title=True)
            self.assertGreater(len(res), 10)

    def test_shape(self):
        mat = gb.Matrix.from_coo([0, 1, 2, 3, 4], [0, 0, 0, 0, 0], [0, 1, 2, 3, 4], nrows=5, ncols=1)
        adapter = matrepr._get_adapter(mat, None)
        self.assertEqual((5, 1), adapter.get_shape())

    def test_contents(self):
        mat = generate_fixed_value(10, 10)
        gb_mat = gb.io.from_scipy_sparse(mat)
        res = to_html(gb_mat, notebook=False, max_rows=20, max_cols=20, title=True, indices=True)
        for value in mat.data:
            self.assertIn(f"<td>{value}</td>", res)

    def test_truncate(self):
        mat = generate_fixed_value(20, 20)
        gb_mat = gb.io.from_scipy_sparse(mat)

        for after_dots, expected_count in [
            (0, 25),  # 5*5
            (1, 25),  # 4*4 + 4 + 4 + 1
            (2, 25),  # 3*3 + 2*3 + 3*2 + 2*2
        ]:
            res = to_html(gb_mat, notebook=False, max_rows=6, max_cols=6, num_after_dots=after_dots)
            count = 0
            for value in mat.data:
                if f"<td>{value}</td>" in res:
                    count += 1
            self.assertEqual(expected_count, count)


@unittest.skipIf(not have_gb, "python-graphblas not installed")
class GraphBLASVectorTests(unittest.TestCase):
    def setUp(self):
        self.vecs = [
            gb.Vector.from_coo([0, 3, 4, 6], [12.1, -5.4, 2.9, 2.2], size=8)
        ]

    def test_no_crash(self):
        for vec in self.vecs:
            res = to_html(vec, title=True)
            self.assertGreater(len(res), 10)

            res = to_latex(vec, title=True)
            self.assertGreater(len(res), 10)

    def test_shape(self):
        vec = gb.Vector.from_coo([0, 3, 4, 6], [12.1, -5.4, 2.9, 2.2], size=8)
        adapter = matrepr._get_adapter(vec, None)
        self.assertEqual((8,), adapter.get_shape())

    def test_contents(self):
        values = [1000, 1001, 1002, 1003, 1004]
        vec = gb.Vector.from_coo([0, 1, 2, 3, 4], values, size=10)
        res = to_html(vec, notebook=False, max_rows=20, max_cols=20, title=True, indices=True)
        for value in values:
            self.assertIn(f"<td>{value}</td>", res)

    def test_truncate(self):
        values = [1000, 1001, 1002, 1003, 1004]
        vec = gb.Vector.from_coo([0, 1, 2, 3, 4], values, size=5)
        res = to_html(vec, notebook=False, max_rows=3, max_cols=3, num_after_dots=1, title=True, indices=True)
        for value in [1000, 1004]:
            self.assertIn(f"<td>{value}</td>", res)


@unittest.skipIf(not have_gb, "python-graphblas not installed")
class PatchGraphBLASTests(unittest.TestCase):
    def test_patch_graphblas(self):
        mat = gb.Matrix.from_coo([0, 1], [0, 1], [111, 222], nrows=5, ncols=5),
        vec = gb.Vector.from_coo([0, 1], [111, 222], size=8)

        # noinspection PyUnresolvedReferences
        import matrepr.patch.graphblas

        res = repr(mat)
        self.assertIn("222", res)
        self.assertIn("┌", res)  # a character used by MatRepr

        res = repr(vec)
        self.assertIn("222", res)
        self.assertIn("[", res)  # a character used by MatRepr

    def test_type_registration(self):
        from matrepr.adapters.graphblas_driver import GraphBLASDriver
        reg_types = [
            typ for typ, _ in GraphBLASDriver.get_supported_types()
        ]

        for tp in [gb.Matrix, gb.Vector]:
            self.assertIn(f"{tp.__module__}.{tp.__name__}", reg_types)


if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

from typing import Any, Dict, Tuple, Optional
from tabulate import tabulate, TableFormat, Line, DataRow

from . import params, _get_adapter
from .adapters import to_trunc
from .list_converter import ListConverter

row_vector_table_format = TableFormat(
    lineabove=None,
    linebelowheader=None,
    linebetweenrows=None,
    linebelow=None,
    headerrow=DataRow(" ", "  ", " "),
    datarow=DataRow("[", "  ", "]"),
    padding=0,
    with_header_hide=None,
)
"""
Renders a table that looks like a horizontal list, supporting a column index header.
"""

row_vector_comma_table_format = TableFormat(
    lineabove=None,
    linebelowheader=None,
    linebetweenrows=None,
    linebelow=None,
    headerrow=DataRow(" ", "  ", " "),
    datarow=DataRow("[", ", ", "]"),
    padding=0,
    with_header_hide=None,
)
"""
Renders a table that looks like a horizontal list, using comma separators instead of a header.
"""


matrix_table_format = TableFormat(
    lineabove=Line("┌", " ", " ", "┐"),
    linebelowheader=Line("┌", " ", " ", "┐"),
    linebetweenrows=None,
    linebelow=Line("└", " ", " ", "┘"),
    headerrow=DataRow(" ", " ", " "),
    datarow=DataRow("│", " ", "│"),
    padding=1,
    with_header_hide=["lineabove"],
)
"""
Renders table that looks like a LaTeX bmatrix.

Renders column indices correctly, but not row indices.
"""

matrix_format_patched_indices = TableFormat(
    lineabove=None,
    linebelowheader=Line("", " ", "┌", "┐"),
    linebetweenrows=None,
    linebelow=Line("", " ", "└", "┘"),
    headerrow=DataRow("", " ", ""),
    datarow=DataRow("", "│", "│"),
    padding=1,
    with_header_hide=None,
)
"""
A corrupted version of :ref:`matrix_table_format` that, when processed using :func:`_tabulate_sep_to_border`,
yields the same beautiful bmatrix but with indices nicely outside the brackets.
"""

tensor_table_format = TableFormat(
    lineabove=None,
    linebelowheader=None,
    linebetweenrows=None,
    linebelow=None,
    headerrow=DataRow(" ", "  ", " "),
    datarow=DataRow("(", ", ", ")"),
    padding=0,
    with_header_hide=None,
)
"""
Renders a table that looks like a horizontal tuple, using comma separators.
"""


tensor_format_patched_indices = TableFormat(
    lineabove=None,
    linebelowheader=None,
    linebetweenrows=None,
    linebelow=None,
    headerrow=DataRow("", " ", ""),
    datarow=DataRow("", "(", ")"),
    padding=1,
    with_header_hide=None,
)
"""
Renders a table that looks like a horizontal tuple, using comma separators.
"""


def _tabulate_sep_to_border(s: str, tf: TableFormat) -> str:
    lines = s.split("\n")

    # Must correct below-header row, data rows, bottom row.
    for row in range(1, len(lines)):
        line = lines[row]
        save_inds = []

        if tf.linebelowheader is not None and tf.linebelow is not None:
            if row == 1:
                # below-header row
                needle = tf.linebelowheader.sep
            elif row < len(lines) - 1:
                # data row uses | separator which appears at the very end too
                needle = tf.datarow.sep
                save_inds.append(-1)
            else:
                # bottom row
                needle = tf.linebelow.sep
        else:
            needle = tf.datarow.sep

        # The leftmost separator is actually the matrix border, so save that one
        save_inds.append(line.find(needle))

        # convert all but the one or two saved copies of the separator to spaces
        line_chars = list(line.replace(needle, " "))
        for i in save_inds:
            line_chars[i] = needle

        lines[row] = "".join(line_chars)
    return "\n".join(lines)


def _labels_to_str(labels):
    for label in labels:
        yield "" if label is None else str(label)


def max_line_width(s: str):
    return max(len(line) for line in s.split("\n"))


def _to_tabulate_args(mat: Any, is_tensor=False, **kwargs) -> Tuple[Dict, Optional[TableFormat]]:
    """
    Creates arguments for :func:`~tabulate.tabulate` to format the matrix.

    Note that tabulate is not expressive enough to render a matrix with nice row indices. Tabulate does not distinguish
    between the index column and data columns. Therefore, index output renders a corrupted output that is then patched
    using :func:`_tabulate_sep_to_border`.

    :param mat: A supported matrix.
    :param kwargs: Any argument in :class:`MatReprParams`, or any argument supported by :func:`~tabulate.tabulate`.
    :return: the argument dictionary and an object. If the object is not None then :func:`_tabulate_sep_to_border`
             must be called with the tabulate output and this object.
    """
    options = params.get(**kwargs)
    adapter = _get_adapter(mat, options)

    # collect the data
    conv = ListConverter(**options.to_kwargs())
    data, row_labels, col_labels = conv.to_lists_and_labels(adapter, is_1d_ok=False)

    # determine the table format
    patch_required = None
    if "tablefmt" in kwargs:
        matrix_format = kwargs["tablefmt"]
    elif not row_labels:
        # horizontal vector
        matrix_format = row_vector_table_format if options.indices else row_vector_comma_table_format
    elif options.indices and row_labels:
        # matrix with row and column indices
        if is_tensor:
            matrix_format = tensor_format_patched_indices
            patch_required = tensor_format_patched_indices
        else:
            matrix_format = matrix_format_patched_indices
            patch_required = matrix_format_patched_indices
    else:
        # matrix without row indices (column indices ok)
        if is_tensor:
            matrix_format = tensor_table_format
        else:
            matrix_format = matrix_table_format

    # collect floatfmt argument
    if "floatfmt" in kwargs and isinstance(kwargs["floatfmt"], str):
        floatfmt = kwargs["floatfmt"]
    elif isinstance(options.floatfmt, str):
        floatfmt = options.floatfmt
    else:
        floatfmt = f".{options.precision}g"

    # base arguments
    tab_args = {
        "tabular_data": data,
        "headers": _labels_to_str(col_labels) if options.indices else (),
        "showindex": _labels_to_str(row_labels) if options.indices and row_labels else False,
        "tablefmt": matrix_format,
        "floatfmt": floatfmt,
    }

    # colalign
    if "colalign" not in kwargs and options.indices:
        tab_args["colalign"] = (["right"] if tab_args["showindex"] else []) + ["center"] * len(col_labels)

    # forward any remaining arguments
    for arg in ["intfmt", "numalign", "stralign", "missingval", "disable_numparse",
                "colalign", "maxcolwidths", "rowalign", "maxheadercolwidths"]:
        if arg in kwargs:
            tab_args[arg] = kwargs[arg]

    return tab_args, patch_required


def to_tabulate(mat: Any, **kwargs) -> str:
    """
    Format a matrix to string using :func:`~tabulate.tabulate`.

    Supports argument passthrough.

    :param mat: A supported matrix.
    :param kwargs: Any argument in :class:`MatReprParams`, or any argument supported by :func:`~tabulate.tabulate`.
    :return: output of :func:`~tabulate.tabulate`
    """
    options = params.get(**kwargs)
    adapter = _get_adapter(mat, options)

    if options.width_str:
        txt_max_cols = max(1, int(options.width_str / 3))  # minimum 3 chars per column (if empty)
        cols = min(options.max_cols, txt_max_cols)
        trunc = to_trunc(adapter, options.max_rows, cols, options.num_after_dots)

        args, patch_tf = _to_tabulate_args(trunc, is_tensor=adapter.is_tensor(), **kwargs)
        ret = tabulate(**args)
        ret_width = max_line_width(ret)

        while ret_width > options.width_str and cols > 1:
            # too wide
            cols = trunc.drop_column()
            args, patch_tf = _to_tabulate_args(trunc, is_tensor=adapter.is_tensor(), **kwargs)
            ret = tabulate(**args)
            ret_width = max_line_width(ret)

    else:
        args, patch_tf = _to_tabulate_args(adapter, is_tensor=adapter.is_tensor(), **kwargs)
        ret = tabulate(**args)

    if patch_tf:
        ret = _tabulate_sep_to_border(ret, patch_tf)
    return ret



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

import unittest

import matrepr


class JupyterExtensionImportTests(unittest.TestCase):
    def test_import_jupyter(self):
        matrepr.load_ipython_extension(None)
        matrepr.unload_ipython_extension(None)
        self.assertEqual(True, True)  # did not raise

    def test_import_jupyter_html(self):
        import matrepr.html
        matrepr.html.load_ipython_extension(None)
        matrepr.html.unload_ipython_extension(None)
        self.assertEqual(True, True)  # did not raise

    def test_import_jupyter_latex(self):
        import matrepr.latex
        matrepr.latex.load_ipython_extension(None)
        matrepr.latex.unload_ipython_extension(None)
        self.assertEqual(True, True)  # did not raise


class BasicTests(unittest.TestCase):
    def test_no_crash_mdisplay(self):
        try:
            import IPython
        except ImportError:
            self.skipTest("no Jupyter")
            return

        mat = [[1, 2], [1003, 1004]]
        matrepr.mdisplay(mat, "html")
        matrepr.mdisplay(mat, "latex")
        matrepr.mdisplay(mat, "str")

        with self.assertRaises(ValueError):
            matrepr.mdisplay(mat, "foobar")

    def test_no_rash_mprint(self):
        mat = [[1000, 1001]]
        matrepr.mprint(mat)
        self.assertEqual(True, True)  # did not raise

    def test_adaptation_errors(self):
        with self.assertRaises(AttributeError):
            matrepr._get_adapter(set(), None)

    def test_arguments(self):
        mat = [[1, 2], [1003, 1004]]
        with self.assertRaises(ValueError):
            matrepr.to_html(mat, cell_align="foobar")

        res = matrepr.to_latex(mat, title="TITLE", title_latex=None)
        self.assertIn("TITLE", res)

        res = matrepr.to_latex(mat, title=None, title_latex="TITLE")
        self.assertIn("TITLE", res)


if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

import unittest
import time

import numpy.random
import scipy.sparse

from matrepr import to_html

numpy.random.seed(123)


class PerformanceTests(unittest.TestCase):
    def test_to_html_speed(self):
        # warmup, just in case
        r = scipy.sparse.random(100, 100, density=0.2, format="csr")
        to_html(r)

        # Verify that to_html() is fast.
        # Using somewhat small dimensions to keep the test snappy.
        a = time.time()
        randmat = scipy.sparse.random(1000, 1000, density=0.2, format="csr")
        b = time.time()
        to_html(randmat)
        c = time.time()

        self.assertGreater((b-a), (c-b))


if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

from typing import Any, Iterable, Tuple

import scipy.sparse

from . import describe, MatrixAdapterRow, MatrixAdapterCol, MatrixAdapterCoo


class SciPyAdapter:
    def __init__(self, mat):
        self.mat = mat

    def get_shape(self) -> tuple:
        return self.mat.shape

    def describe(self) -> str:
        format_name = self.mat.getformat()

        return describe(shape=self.mat.shape, nnz=self.mat.nnz, nz_type=self.mat.dtype,
                        notes=f"{format_name}")


class SciPyCSRAdapter(SciPyAdapter, MatrixAdapterRow):
    def __init__(self, mat: scipy.sparse.csr_matrix):
        super().__init__(mat)

    def get_row(self, row_idx: int, col_range: Tuple[int, int]) -> Iterable[Tuple[int, Any]]:
        start = self.mat.indptr[row_idx]
        end = self.mat.indptr[row_idx + 1]

        indices = self.mat.indices[start:end]
        data = self.mat.data[start:end]

        mask = (col_range[0] <= indices) & (indices < col_range[1])
        return zip(indices[mask], data[mask])


class SciPyCSCAdapter(SciPyAdapter, MatrixAdapterCol):
    def __init__(self, mat: scipy.sparse.csc_matrix):
        super().__init__(mat)

    def get_col(self, col_idx: int, row_range: Tuple[int, int]) -> Iterable[Tuple[int, Any]]:
        start = self.mat.indptr[col_idx]
        end = self.mat.indptr[col_idx + 1]

        indices = self.mat.indices[start:end]
        data = self.mat.data[start:end]

        mask = (row_range[0] <= indices) & (indices < row_range[1])
        return zip(indices[mask], data[mask])


class SciPyCOOAdapter(SciPyAdapter, MatrixAdapterCoo):
    def __init__(self, mat: scipy.sparse.coo_matrix):
        super().__init__(mat)

    def get_coo(self, row_range: Tuple[int, int], col_range: Tuple[int, int]) -> Iterable[Tuple[int, int, Any]]:
        mask = (self.mat.row >= row_range[0]) & (self.mat.row < row_range[1]) & \
               (self.mat.col >= col_range[0]) & (self.mat.col < col_range[1])

        return zip(self.mat.row[mask], self.mat.col[mask], self.mat.data[mask])



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

import unittest

import html5lib
import numpy.random
import scipy.sparse

from matrepr import to_html

numpy.random.seed(123)


def generate_fixed_value(m, n):
    row_factor = 10**(1+len(str(n)))
    nnz = m*n
    rows, cols, data = [1] * nnz, [1] * nnz, [1] * nnz
    for i in range(nnz):
        r = int(i / n)
        c = i % n
        rows[i] = r
        cols[i] = c
        data[i] = (r+1)*row_factor + c

    return scipy.sparse.coo_matrix((data, (rows, cols)), shape=(m, n), dtype='int64')


class ToHTMLTests(unittest.TestCase):
    def setUp(self):
        self.mats = [
            scipy.sparse.random(10, 10, density=0.4),
            scipy.sparse.random(5, 10, density=0.4),
            scipy.sparse.random(5, 1, density=0.4),
            scipy.sparse.coo_matrix(([], ([], [])), shape=(10, 10)),
            generate_fixed_value(10, 10)
        ]

    def test_basic_validate(self):
        """
        A best-effort validation against silly mistakes.
        """
        html5parser = html5lib.HTMLParser(strict=True)

        for mat in self.mats:
            for to_html_args in [
                dict(notebook=False, max_rows=20, max_cols=20, title=False, indices=False),
                dict(notebook=True, max_rows=20, max_cols=20, title=True, indices=True),
                dict(notebook=True, max_rows=4, max_cols=4, title=True, indices=True),
            ]:
                res = to_html(mat, **to_html_args)
                html5parser.parseFragment(res)
                self.assertEqual(True, True)  # did not raise

    def test_contents(self):
        mat = generate_fixed_value(10, 10)
        res = to_html(mat, notebook=False, max_rows=20, max_cols=20, title=True, indices=True)
        for value in mat.data:
            self.assertIn(f"<td>{value}</td>", res)

    def test_truncate(self):
        for fmt in ["csr", "csc", "coo"]:
            mat = generate_fixed_value(20, 20).asformat(fmt)

            for after_dots, expected_count in [
                (0, 25),  # 5*5
                (1, 25),  # 4*4 + 4 + 4 + 1
                (2, 25),  # 3*3 + 2*3 + 3*2 + 2*2
            ]:
                res = to_html(mat, notebook=False, max_rows=6, max_cols=6, num_after_dots=after_dots)
                count = 0
                for value in mat.data:
                    if f"<td>{value}</td>" in res:
                        count += 1
                self.assertEqual(expected_count, count)

    def test_dupes(self):
        row = [0, 0, 0]
        col = [0, 0, 0]
        val = [101, 202, 303]
        mat = scipy.sparse.coo_matrix((val, (row, col)), shape=(5, 5))

        res = to_html(mat, notebook=False, max_rows=6, max_cols=6)

        count = 0
        for value in val:
            if str(value) in res:
                count += 1
        self.assertEqual(len(val), count)

    def test_precision(self):
        f = 0.123456789
        mat = scipy.sparse.coo_matrix(([f], ([0], [0])), shape=(1, 1))

        # default precision
        res = to_html(mat)
        self.assertIn("<td>0.1235</td>", res)

        # explicit format string
        res = to_html(mat, floatfmt=".2g")
        self.assertIn("<td>0.12</td>", res)

        # explicit precision number
        res = to_html(mat, precision=6)
        self.assertIn("<td>0.123457</td>", res)

    def test_title(self):
        mat = generate_fixed_value(4, 4)
        off = to_html(mat, title=False)
        self.assertNotIn("elements", off)

        on = to_html(mat, title=True)
        self.assertIn(f"{mat.nnz} 'int64' elements", on)

        title = "test title"
        custom = to_html(mat, title=title)
        self.assertIn(title, custom)


if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

import unittest

import numpy.random
import scipy.sparse

from matrepr import to_latex
from .test_html import generate_fixed_value

numpy.random.seed(123)


class ToLatexTests(unittest.TestCase):
    def setUp(self):
        self.mats = [
            scipy.sparse.random(10, 10, density=0.4),
            scipy.sparse.random(5, 10, density=0.4),
            scipy.sparse.random(5, 1, density=0.4),
            scipy.sparse.coo_matrix(([], ([], [])), shape=(10, 10)),
            generate_fixed_value(10, 10)
        ]

    def validate_latex(self, s):
        self.assertGreater(len(s), 0)

    def test_basic_validate(self):
        for mat in self.mats:
            for to_latex_args in [
                dict(max_rows=20, max_cols=20, title=False),
                dict(max_rows=20, max_cols=20, title=True),
                dict(max_rows=20, max_cols=20, title=False, indices=True),
                dict(max_rows=20, max_cols=20, title=True, indices=True),
                dict(max_rows=4, max_cols=4, title=True, indices=True),
            ]:
                res = to_latex(mat, **to_latex_args)
                self.validate_latex(res)

    def test_dupes(self):
        row = [0, 0, 0]
        col = [0, 0, 0]
        val = [101, 202, 303]
        mat = scipy.sparse.coo_matrix((val, (row, col)), shape=(5, 5))

        res = to_latex(mat, notebook=False, max_rows=6, max_cols=6)

        count = 0
        for value in val:
            if str(value) in res:
                count += 1
        self.assertEqual(len(val), count)

    def test_title(self):
        mat = generate_fixed_value(4, 4)
        off = to_latex(mat, title=False)
        self.assertNotIn("elements", off)

        on = to_latex(mat, title=True)
        self.assertIn(f"{mat.nnz} `int64' elements", on)

        title = "test title"
        custom = to_latex(mat, title=title)
        self.assertIn(title, custom)

    def test_precision(self):
        f = 0.123456789
        mat = scipy.sparse.coo_matrix(([f], ([0], [0])), shape=(1, 1))

        # default precision
        res = to_latex(mat)
        self.assertIn("0.1235", res)

        # explicit format string
        res = to_latex(mat, floatfmt=".2g")
        self.assertIn("0.12", res)

        # explicit precision number
        res = to_latex(mat, precision=6)
        self.assertIn("0.123457", res)


if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

import unittest

try:
    import sparse
except ImportError:
    sparse = None

from matrepr import to_html, to_latex, to_str
from .test_html import generate_fixed_value

import numpy.random
numpy.random.seed(123)


@unittest.skipIf(sparse is None, "pydata/sparse not installed")
class PyDataSparseTests(unittest.TestCase):
    def setUp(self):
        self.mats = [
            sparse.COO([], shape=(0,)),
            sparse.COO(coords=[1, 4], data=[11, 44], shape=(10,)),
            sparse.COO([], shape=(0, 0)),
            sparse.COO([], shape=(10, 10)),
            sparse.random((10, 10), density=0.4),
            sparse.COO.from_scipy_sparse(generate_fixed_value(10, 10)),
            sparse.COO(coords=[[0, 0], [0, 0]], data=[111, 222], shape=(13, 13)),
            sparse.COO(coords=[[0, 1], [3, 2], [1, 3]], data=[111, 222], shape=(5, 5, 5)),
        ]

        self.types = [
            sparse.COO,
            sparse.DOK,
            sparse.GCXS,
        ]

    def test_type_registration(self):
        from matrepr.adapters.sparse_driver import PyDataSparseDriver
        reg_types = [
            typ for typ, _ in PyDataSparseDriver.get_supported_types()
        ]

        for tp in self.types:
            self.assertIn(f"{tp.__module__}.{tp.__name__}", reg_types)

    def test_no_crash(self):
        for mat in self.mats:
            res = to_html(mat)
            self.assertGreater(len(res), 10)

            res = to_latex(mat)
            self.assertGreater(len(res), 10)

            res = to_str(mat)
            self.assertGreater(len(res), 10)

    def test_formats(self):
        to_html_args = dict(notebook=False, max_rows=20, max_cols=20, title=False)

        expected = [to_html(m, **to_html_args) for m in self.mats]

        for tp in self.types:
            for i, source_mat in enumerate(self.mats):
                if 13 in source_mat.shape:
                    # this matrix has dupes, which act differently in different formats
                    continue
                mat = tp(source_mat)
                res = to_html(mat, **to_html_args)
                self.assertEqual(expected[i], res)

    def test_contents_1d(self):
        values = [1000, 1001, 1002, 1003, 1004]
        vec = sparse.COO([0, 1, 2, 3, 4], data=values, shape=(10,))
        res = to_html(vec, notebook=False, max_rows=20, max_cols=20, title=True, indices=True)
        for value in values:
            self.assertIn(f"<td>{value}</td>", res)

    def test_truncate_1d(self):
        values = [1000, 1001, 1002, 1003, 1009]
        vec = sparse.COO([0, 1, 2, 3, 9], data=values, shape=(10,))
        res = to_html(vec, notebook=False, max_rows=3, max_cols=3, num_after_dots=1, title=True, indices=True)
        for value in [1000, 1009]:
            self.assertIn(f"<td>{value}</td>", res)

    def test_contents_2d(self):
        mat = generate_fixed_value(10, 10)
        sparse_mat = sparse.COO.from_scipy_sparse(mat)
        res = to_html(sparse_mat, notebook=False, max_rows=20, max_cols=20, title=True, indices=True)
        for value in mat.data:
            self.assertIn(f"<td>{value}</td>", res)

    def test_truncate_2d(self):
        mat = generate_fixed_value(20, 20)
        sparse_mat = sparse.COO.from_scipy_sparse(mat)

        for after_dots, expected_count in [
            (0, 25),  # 5*5
            (1, 25),  # 4*4 + 4 + 4 + 1
            (2, 25),  # 3*3 + 2*3 + 3*2 + 2*2
        ]:
            res = to_html(sparse_mat, notebook=False, max_rows=6, max_cols=6, num_after_dots=after_dots)
            count = 0
            for value in mat.data:
                if f"<td>{value}</td>" in res:
                    count += 1
            self.assertEqual(expected_count, count)

    def test_contents_3d(self):
        values = [111, 222]
        mat = sparse.COO(coords=[[0, 1], [3, 2], [1, 3]], data=values, shape=(5, 5, 5))
        res = to_html(mat, notebook=False, max_rows=20, max_cols=20, title=True, indices=True)
        res_str = to_str(mat)
        for value in values:
            self.assertIn(f"<td>{value}</td>", res)
            self.assertIn(f"{value}", res_str)
        self.assertIn(f"<th>val</th>", res)
        self.assertIn(f"val", res_str)

    def test_truncate_3d(self):
        values = [111, 222]
        mat = sparse.COO(coords=[[0, 1], [3, 2], [1, 3]], data=values, shape=(5, 5, 5))

        res = to_html(mat, notebook=False, max_rows=30, max_cols=3, num_after_dots=1)
        res_str = to_str(mat, max_rows=30, max_cols=3, num_after_dots=1)
        count = count_str = 0
        for value in values:
            if f"<td>{value}</td>" in res:
                count += 1
            if f"{value}" in res_str:
                count_str += 1
        self.assertEqual(len(values), count)
        self.assertEqual(len(values), count_str)

        self.assertIn(f"<th>val</th>", res)
        self.assertIn(f"val", res_str)

    def test_patch_sparse(self):
        source_mat = sparse.COO(coords=[1, 4, 6], data=[11, 44, 222], shape=(10,))

        # noinspection PyUnresolvedReferences
        import matrepr.patch.sparse

        for tp in self.types:
            mat = tp(source_mat)
            res = repr(mat)
            self.assertIn("222", res)
            self.assertIn("[", res)  # a character used by MatRepr


if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

from typing import Any, Iterable, Tuple

from . import describe, MatrixAdapterRow


class NumpyArrayAdapter(MatrixAdapterRow):
    def __init__(self, mat):
        self.mat = mat
        self.is_vec = len(mat.shape) < 2

    def get_shape(self) -> tuple:
        # present at most 2D, higher dimensions will be handled as nested arrays
        if self.is_vec:
            return (self.mat.shape[0],)
        else:
            return self.mat.shape[0:2]

    def describe(self) -> str:
        return describe(shape=self.mat.shape,
                        nnz=self.mat.size, nz_type=self.mat.dtype,
                        notes="array")

    def get_row(self, row_idx: int, col_range: Tuple[int, int]) -> Iterable[Any]:
        row = self.mat if self.is_vec else self.mat[row_idx]
        return enumerate(row[col_range[0]:col_range[1]], start=col_range[0])



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

from abc import ABC, abstractmethod
from typing import Any, List, Iterable, Optional, Tuple


def describe(shape: tuple = None, nnz: int = None, nz_type=None, notes: str = None) -> str:
    """
    Create a simple description string from potentially interesting pieces of metadata.
    """
    parts = []
    by = chr(215)  # ×
    if len(shape) == 1:
        parts.append(f"length={shape[0]}")
    elif len(shape) == 2:
        parts.append(f"{shape[0]}{by}{shape[1]}")
    else:
        parts.append(f"shape=({','.join(str(d) for d in shape)})")

    if nnz is not None:
        dtype_str = f" '{str(nz_type)}'" if nz_type else ""
        parts.append(f"{nnz}{dtype_str} elements")

    if notes:
        parts.append(notes)

    return ", ".join(parts)


class DupeList(list):
    """
    A list but a different type to distinguish between a list in the original data
    and a list created to handle duplicate entries.
    """

    def __init__(self, iterable):
        super().__init__(iterable)


class MatrixAdapter(ABC):
    @abstractmethod
    def describe(self) -> str:
        """
        Return a human-readable description of the matrix, usable as a title.
        """

    @abstractmethod
    def get_shape(self) -> tuple:
        """
        Return the shape of the matrix.
        """

    def get_row_labels(self) -> Optional[Iterable[Optional[int]]]:
        return self.row_labels if hasattr(self, "row_labels") else range(self.get_shape()[0])

    def get_col_labels(self) -> Iterable[Optional[Any]]:
        return self.col_labels if hasattr(self, "col_labels") else range(self.get_shape()[1])

    # noinspection PyMethodMayBeStatic
    def is_tensor(self) -> bool:
        return False


class MatrixAdapterRow(MatrixAdapter):
    @abstractmethod
    def get_row(self, row_idx: int, col_range: Tuple[int, int]) -> Iterable[Tuple[int, Any]]:
        """
        Extract a portion of single row from the matrix, as sparse tuples.

        :param row_idx: index of row to fetch
        :param col_range: half-open range of column indices to return
        :return: an iterable of `(index, value)` tuples
        """

    def get_dense_row(self, row_idx: int, col_range: Tuple[int, int]) -> Iterable[Any]:
        """
        Extract a portion of single row from the matrix, as a dense array.

        :param row_idx: index of row to fetch
        :param col_range: half-open range of columns to return
        :return: an iterable of length `col_range[1] - col_range[0]`
        """
        ret: List[Any] = [None] * (col_range[1] - col_range[0])
        for idx, value in self.get_row(row_idx, col_range):
            idx = idx - col_range[0]
            if ret[idx] is None:
                ret[idx] = value
            elif isinstance(ret[idx], DupeList):
                ret[idx].append(value)
            else:
                ret[idx] = DupeList([ret[idx], value])
        return ret


class MatrixAdapterCol(MatrixAdapter):
    @abstractmethod
    def get_col(self, col_idx: int, row_range: Tuple[int, int]) -> Iterable[Tuple[int, Any]]:
        """
        Extract a portion of single column from the matrix, as sparse tuples.

        :param col_idx: index of column to fetch
        :param row_range: half-open range of rows to return
        :return: an iterable of `(index, value)` tuples
        """


class MatrixAdapterCoo(MatrixAdapter):
    @abstractmethod
    def get_coo(self, row_range: Tuple[int, int], col_range: Tuple[int, int]) -> Iterable[Tuple[int, int, Any]]:
        """
        Extract a portion of the matrix, as sparse tuples.

        :param row_range: half-open range of rows to return
        :param col_range: half-open range of columns to return
        :return: an iterable of `(row_index, column_index, value)` tuples
        """


class TensorAdapterCooRow(MatrixAdapterRow):
    """
    Present an n-dimensional tensor as a 2D table where each row is a coordinate tuple of length n+1:
    coordinates and value.
    """
    @abstractmethod
    def get_shape(self) -> Tuple[int, int]:
        """
        Return the shape of the tabular view of the tensor.

        :returns: a tuple: (nnz, number of dimensions+1)
        """

    @abstractmethod
    def get_dense_row(self, row_idx: int, col_range: Tuple[int, int]) -> Iterable[Any]:
        """
        Extract a single tuple from the tensor, as a dense array.

        :param row_idx: index of tensor tuple to fetch
        :param col_range: half-open range of dimensions to return
        :return: an iterable of length `col_range[1] - col_range[0]`
        """

    def get_col_labels(self) -> Iterable[Optional[Any]]:
        if hasattr(self, "col_labels"):
            return self.col_labels

        ret = list(range(self.get_shape()[1]))
        if ret:
            ret[-1] = "val"
        return ret

    def get_row(self, row_idx: int, col_range: Tuple[int, int]) -> Iterable[Tuple[int, Any]]:
        return enumerate(self.get_dense_row(row_idx, col_range), start=col_range[0])

    # noinspection PyMethodMayBeStatic
    def is_tensor(self) -> bool:
        return True


class Driver(ABC):
    @staticmethod
    @abstractmethod
    def get_supported_types() -> Iterable[Tuple[str, bool]]:
        """
        Declares the types that this :class:`Driver` supports, and whether they should be registered with Jupyter.

        Does not import the modules that it supports.

        :rtype: (str, bool)
        :returns: An iterable of `(type_as_str, should_register_with_Jupyter_bool)` tuples.
        """

    @staticmethod
    @abstractmethod
    def adapt(mat: Any) -> Optional[MatrixAdapter]:
        """
        Return a :class:`MatrixAdapter` for a supported matrix.
        """


class Truncated2DMatrix(MatrixAdapterRow):
    """
    An intermediary class used to present a dense view of an adapted matrix.

    - Convert any adapter to a :class:`MatrixAdapterRow`.
    - If a matrix is too large then supports showing just the corners with ellipses designating the truncated portions.
    """
    def __init__(self, orig_shape: Tuple[int, int], display_shape: Tuple[int, int], num_after_dots=2,
                 row_labels=None, col_labels=None, description=None):
        self.show_row_labels = len(orig_shape) != 1
        if len(orig_shape) == 1:
            orig_shape = (1, orig_shape[0])
        elif len(orig_shape) > 2:
            raise ValueError
        self.orig_shape = orig_shape
        self.display_shape = [min(orig_shape[0], display_shape[0]), min(orig_shape[1], display_shape[1])]
        self.nrows, self.ncols = self.display_shape
        self.num_after_dots = num_after_dots
        self.elements: List[Any] = [[None] * self.ncols for _ in range(self.nrows)]
        self.orig_row_labels = row_labels
        self.orig_col_labels = col_labels
        self.description = description

        self.dot_col = None
        self.dot_row = None

        if self.display_shape[0] < self.orig_shape[0]:
            # need to truncate rows
            self.dot_row = self._calc_dots(self.display_shape[0], self.orig_shape[0])

        if self.display_shape[1] < self.orig_shape[1]:
            # need to truncate columns
            self.dot_col = self._calc_dots(self.display_shape[1], self.orig_shape[1])

    def _calc_dots(self, display, orig):
        if display >= orig:
            return orig

        if 0 < self.num_after_dots < 1:
            # fractional
            return int(display * self.num_after_dots)
        else:
            return max(0, display - 1 - int(self.num_after_dots))

    def describe(self) -> str:
        return self.description

    def get_shape(self):
        return self.display_shape

    def get_displayed_row_indices(self) -> List[int]:
        pre_dot_end, post_dot_start = self.get_dot_indices_row()
        return list(range(pre_dot_end)) + list(range(post_dot_start, self.orig_shape[0]))

    def get_displayed_col_indices(self) -> List[int]:
        pre_dot_end, post_dot_start = self.get_dot_indices_col()
        return list(range(pre_dot_end)) + list(range(post_dot_start, self.orig_shape[1]))

    def get_row_labels(self) -> Optional[Iterable[Optional[int]]]:
        if not self.show_row_labels:
            return None

        if self.dot_row is None:
            return self.orig_row_labels if self.orig_row_labels else [str(i) for i in range(self.orig_shape[0])]
        else:
            pre_dot_end, post_dot_start = self.get_dot_indices_row()
            if self.orig_row_labels is None:
                # generate indices
                # noinspection PyTypeChecker
                return [str(i) for i in range(pre_dot_end)] + [None]\
                    + [str(i) for i in range(post_dot_start, self.orig_shape[0])]
            else:
                return self.orig_row_labels[:pre_dot_end] + [None] \
                    + self.orig_row_labels[post_dot_start:self.orig_shape[0]]

    def get_col_labels(self) -> Iterable[Optional[int]]:
        if self.dot_col is None:
            return self.orig_col_labels if self.orig_col_labels else [str(i) for i in range(self.orig_shape[1])]
        else:
            pre_dot_end, post_dot_start = self.get_dot_indices_col()
            if self.orig_col_labels is None:
                # generate indices
                # noinspection PyTypeChecker
                return [str(i) for i in range(pre_dot_end)] + [None]\
                    + [str(i) for i in range(post_dot_start, self.orig_shape[1])]
            else:
                return self.orig_col_labels[:pre_dot_end] + [None] \
                    + self.orig_col_labels[post_dot_start:self.orig_shape[1]]

    def get_row(self, row_idx: int, col_range: Tuple[int, int]) -> Iterable[Tuple[int, Any]]:
        return enumerate(self.get_dense_row(row_idx, col_range), start=col_range[0])

    def get_dense_row(self, row_idx: int, col_range: Tuple[int, int]):
        return self.elements[row_idx][col_range[0]:col_range[1]]

    def set(self, row_idx: int, col_idx: int, value: Any):
        row_idx = int(row_idx)
        col_idx = int(col_idx)

        pre_dot_end, post_dot_start = self.get_dot_indices_row()
        if self.dot_row is not None and row_idx >= pre_dot_end:
            row_idx = self.dot_row + 1 + (row_idx - post_dot_start)
            if row_idx < self.dot_row:
                # within the dots
                return

        pre_dot_end, post_dot_start = self.get_dot_indices_col()
        if self.dot_col is not None and col_idx >= pre_dot_end:
            col_idx = self.dot_col + 1 + (col_idx - post_dot_start)
            if col_idx < self.dot_col:
                # within the dots
                return

        if self.elements[row_idx][col_idx] is None:
            self.elements[row_idx][col_idx] = value
        elif isinstance(self.elements[row_idx][col_idx], DupeList):
            self.elements[row_idx][col_idx].append(value)
        else:
            self.elements[row_idx][col_idx] = DupeList([self.elements[row_idx][col_idx], value])

    def apply_dots(self, dots):
        if self.dot_row is not None:
            for i in range(len(self.elements[self.dot_row])):
                self.elements[self.dot_row][i] = dots["v"]
        if self.dot_col is not None:
            for i in range(len(self.elements)):
                self.elements[i][self.dot_col] = dots["h"]
            if self.dot_row is not None:
                self.elements[self.dot_row][self.dot_col] = dots["d"]

    def get_dot_indices_row(self) -> Tuple[int, int]:
        if self.dot_row is None:
            return self.orig_shape[0], self.orig_shape[0]
        else:
            num_post_dot = self.display_shape[0] - 1 - self.dot_row
            return self.dot_row, self.orig_shape[0] - num_post_dot

    def get_dot_indices_col(self) -> Tuple[int, int]:
        if self.dot_col is None:
            return self.orig_shape[1], self.orig_shape[1]
        else:
            num_post_dot = self.display_shape[1] - 1 - self.dot_col
            return self.dot_col, self.orig_shape[1] - num_post_dot

    def drop_column(self):
        if self.display_shape[1] < 2:
            return self.display_shape[1]

        # determine which column to drop
        old_dot_col = self.dot_col
        self.dot_col = self._calc_dots(self.display_shape[1] - 1, self.orig_shape[1])

        if old_dot_col is None:
            if self.dot_col > 0:
                drop = self.dot_col - 1
            else:
                drop = self.dot_col
        else:
            assert self.dot_col <= old_dot_col
            if self.dot_col == 0:
                drop = self.dot_col
            elif self.dot_col < old_dot_col:
                drop = self.dot_col - 1
            else:
                drop = self.dot_col

        # adjust metadata
        self.display_shape[1] -= 1
        self.ncols -= 1

        # drop
        for row in self.elements:
            row.pop(drop)

        return self.display_shape[1]


def to_trunc(mat: MatrixAdapter, max_rows, max_cols, num_after_dots) -> Truncated2DMatrix:
    """
    Convert an adapted matrix to a :class:`Truncated2DMatrix`.
    """
    if isinstance(mat, Truncated2DMatrix):
        return mat

    shape = mat.get_shape()
    if len(shape) == 1:
        # vector
        nrows = 1
        ncols = mat.get_shape()[0]
    elif len(shape) == 2:
        nrows, ncols = mat.get_shape()
        max_rows = min(max_rows, nrows)
        max_cols = min(max_cols, ncols)
    else:
        raise ValueError("Only 1 or 2 dimensional matrices supported at this time.")

    if nrows == 0 or ncols == 0:
        return Truncated2DMatrix(orig_shape=mat.get_shape(),
                                 display_shape=(max_rows, max_cols),
                                 num_after_dots=0,
                                 description=mat.describe())

    if isinstance(mat, MatrixAdapterCoo):
        if max_rows >= nrows and max_cols >= ncols:
            num_after_dots = 0

        import time
        begin = time.time()
        top_left = list(mat.get_coo(row_range=(0, max_rows), col_range=(0, max_cols)))
        duration = time.time() - begin

        if duration > 0.1:
            # slow to get data, only show the top-left
            num_after_dots = 0

        if num_after_dots == 0:
            trunc = Truncated2DMatrix(orig_shape=mat.get_shape(),
                                      display_shape=(max_rows, max_cols),
                                      num_after_dots=0,
                                      row_labels=mat.row_labels if hasattr(mat, "row_labels") else None,
                                      col_labels=mat.col_labels if hasattr(mat, "col_labels") else None,
                                      description=mat.describe())
            for row, col, val in top_left:
                trunc.set(row, col, val)
        else:
            trunc = Truncated2DMatrix(orig_shape=mat.get_shape(),
                                      display_shape=(max_rows, max_cols),
                                      num_after_dots=num_after_dots,
                                      row_labels=mat.row_labels if hasattr(mat, "row_labels") else None,
                                      col_labels=mat.col_labels if hasattr(mat, "col_labels") else None,
                                      description=mat.describe())
            for row, col, val in top_left:
                trunc.set(row, col, val)

            # fetch the other three quadrants
            dot_row = trunc.dot_row if trunc.dot_row else max_rows
            dot_col = trunc.dot_col if trunc.dot_col else max_cols
            for row_range, col_range in [
                ((0, dot_row), (max(max_cols, dot_col + 1), ncols)),  # top-right
                ((max(max_rows, dot_row + 1), nrows), (0, dot_col)),  # bottom-left
                ((max(max_rows, dot_row + 1), nrows), (max(max_cols, dot_col + 1), ncols)),  # bottom-right
            ]:
                if row_range[1] - row_range[0] <= 0 or col_range[1] - col_range[0] <= 0:
                    continue

                for row, col, val in mat.get_coo(row_range=row_range, col_range=col_range):
                    trunc.set(row, col, val)
        return trunc

    if isinstance(mat, MatrixAdapterRow):
        if isinstance(mat, TensorAdapterCooRow):
            mat.col_labels = mat.get_col_labels()

        trunc = Truncated2DMatrix(orig_shape=mat.get_shape(),
                                  display_shape=(max_rows, max_cols),
                                  num_after_dots=num_after_dots,
                                  row_labels=mat.row_labels if hasattr(mat, "row_labels") else None,
                                  col_labels=mat.col_labels if hasattr(mat, "col_labels") else None,
                                  description=mat.describe())

        pre_dot_end, post_dot_start = trunc.get_dot_indices_col()

        # fetch the pre-dot rows
        rows_to_fetch = [0] if len(shape) == 1 else trunc.get_displayed_row_indices()
        for row_idx in rows_to_fetch:
            if row_idx is None:
                # dots
                continue

            for col_range in [(0, pre_dot_end), (post_dot_start, ncols)]:
                for col_idx, value in mat.get_row(row_idx, col_range=col_range):
                    trunc.set(row_idx, col_idx, value)
        return trunc

    if isinstance(mat, MatrixAdapterCol):
        trunc = Truncated2DMatrix(orig_shape=mat.get_shape(),
                                  display_shape=(max_rows, max_cols),
                                  num_after_dots=num_after_dots,
                                  row_labels=mat.row_labels if hasattr(mat, "row_labels") else None,
                                  col_labels=mat.col_labels if hasattr(mat, "col_labels") else None,
                                  description=mat.describe())

        pre_dot_end, post_dot_start = trunc.get_dot_indices_row()

        # fetch the pre-dot rows
        for col_idx in trunc.get_displayed_col_indices():
            if col_idx is None:
                # dots
                continue

            for row_range in [(0, pre_dot_end), (post_dot_start, nrows)]:
                for row_idx, value in mat.get_col(col_idx, row_range=row_range):
                    trunc.set(row_idx, col_idx, value)
        return trunc

    raise NotImplementedError



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

import html

from .adapters import MatrixAdapter, MatrixAdapterRow, Truncated2DMatrix, to_trunc, DupeList
from .base_formatter import BaseFormatter, unicode_dots


html_dots = {
    "h": "&ctdot;",  # '⋯'
    "v": "&vellip;",  # '⋮'
    "d": "&dtdot;"  # '⋱'
}

unicode_to_html = {
    u: html_dots[k] for k, u in unicode_dots.items()
}


class HTMLTableFormatter(BaseFormatter):
    def __init__(self, max_rows, max_cols, num_after_dots, title, indices=False, cell_align="center",
                 floatfmt=None, **_):
        super().__init__()
        self.max_rows = max_rows
        self.max_cols = max_cols
        self.num_after_dots = num_after_dots
        self.title = title
        self.indices = indices
        self.cell_align = cell_align
        self.floatfmt = floatfmt if floatfmt else lambda f: format(f)
        self.indent_width = 4
        self.center_header = False
        self.table_attributes = {}

    # noinspection PyMethodMayBeStatic
    def _attributes_to_string(self, attributes: dict) -> str:
        if attributes is None:
            return ""

        ret = ' '.join([f'{k}="{v}"' for k, v in attributes.items()])
        if ret:
            ret = " " + ret
        return ret

    def pprint(self, obj, current_indent=0):
        if obj is None:
            return ""

        if isinstance(obj, (int, float)):
            return self.floatfmt(obj)

        if isinstance(obj, complex):
            sign = "-" if obj.imag < 0 else "+"
            r = self.pprint(obj.real).replace("e+", "e")
            c = self.pprint(obj.imag).replace("e+", "e")
            if "e" in r:
                # add space to visually separate the exponent from the complex plus
                sign = f" {sign} "
            return f"{r}{sign}{c}<i>i</i>"

        if isinstance(obj, str):
            if obj in unicode_to_html:
                return unicode_to_html[obj]
            else:
                return html.escape(obj).replace("\n", "<br>")

        if isinstance(obj, DupeList):
            return "<br>".join(self.pprint(sub) for sub in obj)

        from . import _get_adapter
        adapter = _get_adapter(obj, None, unsupported_raise=False)
        if adapter:
            from copy import copy
            fmt = copy(self)
            fmt.lines = []
            fmt.max_rows = max(self.max_rows / 2, 2)
            fmt.max_cols = max(self.max_cols / 2, 2)
            fmt.num_after_dots = 0
            fmt.title = None
            fmt.indices = False
            fmt.table_attributes["style"] = "margin: 0 auto;"
            return "\n" + str(fmt.format(adapter, indent=current_indent + self.indent_width))

        return html.escape(str(obj))

    def _write_matrix(self, mat: MatrixAdapterRow,
                      indent: int = 0):
        if isinstance(mat, Truncated2DMatrix):
            mat.apply_dots(unicode_dots)

        nrows, ncols = mat.get_shape()
        self.write(f"<table{self._attributes_to_string(self.table_attributes)}>", indent=indent)

        body_indent = indent + self.indent_width
        cell_indent = body_indent + self.indent_width

        row_labels = mat.get_row_labels() if self.indices else None
        row_labels = iter(row_labels) if row_labels else None

        # Header
        if self.indices:
            self.write(f'<thead>', indent=body_indent)
            self.write("<tr>", indent=body_indent)
            if row_labels:
                self.write(f"<th></th>", indent=cell_indent)
            for col_label in mat.get_col_labels():
                attr = ' style="text-align: center;"' if self.center_header else ""
                self.write(f"<th{attr}>{self.pprint(col_label)}</th>", indent=cell_indent)
            self.write("</tr>", indent=body_indent)
            self.write("</thead>", indent=body_indent)

        # values
        self.write("<tbody>", indent=body_indent)
        for row_idx in range(nrows):
            self.write("<tr>", body_indent)
            if row_labels:
                self.write(f"<th>{self.pprint(next(row_labels))}</th>", cell_indent)

            col_range = (0, ncols)
            for col_idx, cell in enumerate(mat.get_dense_row(row_idx, col_range=col_range)):
                self.write(f"<td>{self.pprint(cell, cell_indent)}</td>", cell_indent)

            self.write("</tr>", body_indent)

        self.write("</tbody>", indent=body_indent)

        self.write("</table>", indent=indent)

    def format(self, mat: MatrixAdapter, indent: int = 0):
        if self.title:
            title = mat.describe() if self.title is True else self.title
            self.write(f"<p>{title}</p>")

        if not isinstance(mat, MatrixAdapterRow) or \
                len(mat.get_shape()) != 2 or \
                mat.get_shape()[0] > self.max_rows or \
                mat.get_shape()[1] > self.max_cols:
            mat = to_trunc(mat, self.max_rows, self.max_cols, self.num_after_dots)

        self._write_matrix(mat, indent=indent)

        return self


class NotebookHTMLFormatter(HTMLTableFormatter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.wrote_style = False

    def _write_style(self, matrix_class: str, tensor_class: str):
        self.wrote_style = True

        self.write("<style scoped>")

        index_attributes = {
            "font-size": "smaller",
            "vertical-align": "middle"
        }

        border = "solid 2px"

        # for bracket tick marks
        tick_common = {
            "content": '""',
            "width": "4px",
            "position": "absolute",
            "top": 0,
            "bottom": 0,
            "visibility": "visible",
        }
        left_ticks = {
            **tick_common,
            "left": 0,
            "right": "auto",
        }
        right_ticks = {
            **tick_common,
            "left": "auto",
            "right": 0,
        }

        table = f"table.{matrix_class} "
        thead = table + "> thead "
        tbody = table + "> tbody "

        tensor = f"table.{tensor_class} "
        tenhead = tensor + "> thead "
        tenbody = tensor + "> tbody "

        empty_content = r'"\00a0\00a0\00a0"'  # will be doubled

        for tags, attributes in [
            (table, {"border-collapse": "collapse"}),  # already in Jupyter style, but not if copy/pasted elsewhere
            (thead, {"border": "0px"}),
            (tbody + "tr th", {**index_attributes, "text-align": "right"}),  # row indices
            (thead + "tr th", {**index_attributes, "text-align": "center"}),  # column indices
            (tbody + "tr td", {"vertical-align": "middle", "text-align": self.cell_align, "position": "relative"}),
            (tbody + "tr td:first-of-type", {"border-left": border}),  # left border
            (tbody + "tr td:last-of-type", {"border-right": border}),  # right border
            # need two of these because ticks may narrow down one of them
            (tbody + "tr td:empty::before", {"content": empty_content, "visibility": "hidden"}),  # fill empty cells
            (tbody + "tr td:empty::after", {"content": empty_content, "visibility": "hidden"}),  # fill empty cells
            # ticks
            (tbody + "> tr:first-child > td:first-of-type::before", {**left_ticks, "border-top": border}),
            (tbody + "> tr:last-child > td:first-of-type::before", {**left_ticks, "border-bottom": border}),
            (tbody + "> tr:first-child > td:last-of-type::after", {**right_ticks, "border-top": border}),
            (tbody + "> tr:last-child > td:last-of-type::after", {**right_ticks, "border-bottom": border}),

            # tensor styles
            (tenbody + "tr th", {**index_attributes, "text-align": "right"}),  # row indices
            (tenhead + "tr th", {**index_attributes, "text-align": "center"}),  # column indices
            (tenbody + "tr td", {"vertical-align": "middle", "text-align": self.cell_align, "position": "relative"}),
            (tenbody + "tr th:last-of-type", {"border-right": border}),  # left border
        ]:
            self.write(f"{tags} " + '{', indent=self.indent_width)
            for k, v in attributes.items():
                self.write(f"{k}: {v};", indent=2*self.indent_width)
            self.write("}", indent=self.indent_width)
        self.write("</style>")

    def format(self, mat: MatrixAdapter, indent: int = 0):
        matrix_class = "matreprmatrix"
        tensor_class = "matreprtensor"
        table_class = tensor_class if mat.is_tensor() else matrix_class
        self.table_attributes["class"] = table_class

        write_div = not self.wrote_style
        if write_div:
            self.write("<div>")
            self._write_style(matrix_class=matrix_class, tensor_class=tensor_class)
        self.center_header = True
        super().format(mat, indent=indent)
        if write_div:
            self.write("</div>")
        return self



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

import unittest

import numpy as np

from matrepr import to_html, to_latex, to_str

np.random.seed(123)


class NumpyTests(unittest.TestCase):
    def setUp(self):
        self.mats = [
            np.array([]),
            np.array([1, 2, 3, 4]),
            np.array([[1, 2], [1003, 1004]]),
            np.array([
                [[1, 2], [3, 4]],
                [[100, 200], [300, 400]],
            ]),
            np.random.randint(low=0, high=10_000, size=(2, 2, 2, 2))  # 4D
        ]

    def test_no_crash(self):
        for mat in self.mats:
            res = to_str(mat, title=True)
            self.assertGreater(len(res), 10)

            res = to_html(mat, title=True)
            self.assertGreater(len(res), 10)

            res = to_latex(mat, title=True)
            self.assertGreater(len(res), 10)

    def test_contents_html(self):
        for mat in self.mats:
            if len(mat) == 0:
                continue

            res = to_html(mat, notebook=True, max_rows=200, max_cols=200, title=False, indices=False)
            self.assertGreater(len(res), 10)

            for value in np.nditer(mat):
                self.assertIn(f"<td>{value}</td>", res)


if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

from typing import Any, Iterable, Tuple

import graphblas as gb

from . import describe, MatrixAdapterCoo, MatrixAdapterRow


class GraphBLASAdapter:
    def __init__(self, mat):
        self.mat = mat

    def get_shape(self) -> tuple:
        return self.mat.shape

    def get_format(self, is_transposed=False):
        x = self.mat
        try:
            # SS, SuiteSparse-specific: format (ends with "r" or "c"), and is_iso
            fmt = x.ss.format
            if is_transposed:
                fmt = fmt[:-1] + ("c" if fmt[-1] == "r" else "r")
            if x.ss.is_iso:
                return f"{fmt} (iso)"
            return fmt
        except AttributeError:
            return None

    def describe(self) -> str:
        parts = [f"gb.{type(self.mat).__name__}"]

        fmt = self.get_format()
        if fmt:
            parts.append(fmt)

        return describe(shape=self.mat.shape,
                        nnz=self.mat.nvals, nz_type=self.mat.dtype,
                        notes=", ".join(parts))


class GraphBLASMatrixAdapter(GraphBLASAdapter, MatrixAdapterCoo):
    def __init__(self, mat: gb.Matrix):
        super().__init__(mat)

    def get_coo(self, row_range: Tuple[int, int], col_range: Tuple[int, int]) -> Iterable[Tuple[int, int, Any]]:
        ret = self.mat[slice(*row_range), slice(*col_range)]

        rows, cols, vals = ret.to_coo()
        rows += row_range[0]
        cols += col_range[0]
        return zip(rows, cols, vals)


class GraphBLASVectorAdapter(GraphBLASAdapter, MatrixAdapterRow):
    def __init__(self, vec: gb.Vector):
        super().__init__(vec)

    def get_row(self, row_idx: int, col_range: Tuple[int, int]) -> Iterable[Any]:
        assert row_idx == 0

        idx, vals = self.mat[slice(*col_range)].to_coo()
        idx += col_range[0]
        return zip(idx, vals)



# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

"""
Jupyter extension that registers an HTML formatter for supported matrix types.

Usage:
%load_ext matrepr.html
"""


def _repr_html_(mat):
    from . import to_html
    return to_html(mat, notebook=True)


def load_ipython_extension(ipython):
    from . import _register_jupyter_formatter
    _register_jupyter_formatter(ipython, "text/html", _repr_html_)


def unload_ipython_extension(ipython):
    from . import _register_jupyter_formatter
    _register_jupyter_formatter(ipython, "text/html", None)
